﻿#include "cmanagemodbus.h"
#include <string.h>
#include <QDebug>

#ifdef WIN32
typedef int socklen_t;
#endif

/***************************************************************
 * @file       CManageModbus.cpp
 * @author     seer-txj
 * @brief      modbus initialization
 * @param      IP/PORT/debugflag
 * @version    v1
 * @return     null
 * @date       2021/10/6
 **************************************************************/
bool CManageModbus::initModbus(std::string Host_Ip = "127.0.0.1", int port = 502, bool debugging = true)
{
    ctx = modbus_new_tcp(Host_Ip.c_str(), port);
    //modbus_set_debug(ctx, debugging);
    if (ctx == NULL)
    {
        fprintf(stderr, "There was an error allocating the modbus\n");
        throw - 1;
    }
    m_modbusSocket = modbus_tcp_listen(ctx, 1);
    /*设置线圈, 离散输入, 输入寄存器, 保持寄存器个数(数组元素个数))*/
    mapping = modbus_mapping_new(m_numBits, m_numInputBits, m_numInputRegisters, m_numRegisters);
    if (mapping == NULL)
    {
        fprintf(stderr, "Unable to assign mapping：%s\n", modbus_strerror(errno));
        modbus_free(ctx);
        m_initialized = false;
        return false;
    }
    m_initialized = true;
    return true;
}
/***************************************************************
 * @file       CManageModbus.cpp
 * @author     seer-txj
 * @brief      Constructor
 * @version    v1
 * @return     null
 * @date       2021/10/6
 **************************************************************/
CManageModbus::CManageModbus(string host, uint16_t port, bool bServer, bool bRemoveHead, QObject* parent)
    :QObject(parent)
{
    m_bServer = bServer;
    m_bStopThread = false;
    m_bRemoveHead = bRemoveHead;
    initModbus(host, port, false);

    //TODO：
}
/***************************************************************
 * @file       CManageModbus.cpp
 * @author     seer-txj
 * @brief      Destructor
 * @version    v1
 * @return     null
 * @date       2021/10/6
 **************************************************************/
CManageModbus::~CManageModbus()
{
    if(!ctx)
    {
        modbus_mapping_free(mapping);
        modbus_close(ctx);
        modbus_free(ctx);
    }
}
/***************************************************************
 * @file       CManageModbus.cpp
 * @author     seer-txj
 * @brief      loadFromConfigFile
 * @version    v1
 * @return     null
 * @date       2021/10/18
 **************************************************************/
void CManageModbus::loadFromConfigFile()
{
    return;
}
/***************************************************************
 * @file       CManageModbus.cpp
 * @author     seer-txj
 * @brief      run
 * @version    v1
 * @return     null
 * @date       2021/10/18
 **************************************************************/
void CManageModbus::run()
{
    std::thread loop([this]()
    {
        while (true)
        {
            if (m_initialized)
            {
                recieveMessages();
            }
            else
            {
                m_initialized = true;
            }
        }
    });
    loop.detach();
    return;
}

/***************************************************************
 * @file       CManageModbus.cpp
 * @author     seer-txj
 * @brief      modbus_set_slave_id
 * @param      id
 * @version    v1
 * @return     null
 * @date       2021/10/19
 **************************************************************/
bool CManageModbus::modbus_set_slave_id(int id)
{
    int rc = modbus_set_slave(ctx, id);
    if (rc == -1)
    {
        fprintf(stderr, "Invalid slave id\n");
        modbus_free(ctx);
        return false;
    }
    return true;
}


bool CManageModbus::setInputRegisterValue(int registerStartaddress, uint16_t Value)
{
    if (registerStartaddress > (m_numRegisters - 1))
    {
        return false;
    }
    slavemutex.lock();
    mapping->tab_input_registers[registerStartaddress] = Value;
    slavemutex.unlock();
    return true;
}





/***************************************************************
 * @file       CManageModbus.cpp
 * @author     seer-txj
 * @brief      setRegisterValue(设置保存寄存器的值，类型为uint16_t)
 * @param      registerStartaddress(保存寄存器的起始地址)
 * @param      Value(写入到保存寄存器里的值)
 * @version    v1
 * @return     null
 * @date       2021/10/6
 **************************************************************/
bool CManageModbus::setHoldingRegisterValue(int registerStartaddress, uint16_t Value)
{
    if (registerStartaddress > (m_numRegisters - 1))
    {
        return false;
    }
    slavemutex.lock();
    mapping->tab_registers[registerStartaddress] = Value;
    slavemutex.unlock();
    return true;
}
/***************************************************************
 * @file       CManageModbus.cpp
 * @author     seer-txj
 * @brief      getRegisterValue(获取保存寄存器的值)
 * @param      registerStartaddress(保存寄存器的起始地址)
 * @version    v1
 * @return     null
 * @date       2021/10/6
 **************************************************************/
uint16_t CManageModbus::getHoldingRegisterValue(int registerStartaddress)
{
    if (!m_initialized)
    {
        return -1;
    }
    return mapping->tab_registers[registerStartaddress];
}
/***************************************************************
 * @file       CManageModbus.cpp
 * @author     seer-txj
 * @brief      setTab_Input_Bits(设置输入寄存器某一位的值)
 * @param      NumBit(输入寄存器的起始地址)
 * @param      Value(输入寄存器的值)
 * @version    v1
 * @return     null
 * @date       2021/10/8
 **************************************************************/
bool CManageModbus::setTab_Input_Bits(int NumBit, uint8_t Value)
{
    if (NumBit > (m_numInputBits - 1))
    {
        return false;
    }
    slavemutex.lock();
    mapping->tab_input_bits[NumBit] = Value;
    slavemutex.unlock();
    return true;
}
/***************************************************************
 * @file       CManageModbus.cpp
 * @author     seer-txj
 * @brief      getTab_Input_Bits(获取输入寄存器某一位的值)
 * @param      NumBit(输入寄存器相应的bit位)
 * @version    v1
 * @return     null
 * @date       2021/10/8
 **************************************************************/
uint8_t CManageModbus::getTab_Input_Bits(int NumBit)
{
    if (!m_initialized)
    {
        return -1;
    }
    return mapping->tab_input_bits[NumBit];
}
/***************************************************************
 * @file       CManageModbus.cpp
 * @author     seer-txj
 * @brief      setRegisterFloatValue(设置浮点值)
 * @param      (Value：浮点值，registerStartaddress寄存器起始地址)
 * @version    v1
 * @return     null
 * @date       2021/10/8
 **************************************************************/
bool CManageModbus::setHoldingRegisterValue(int registerStartaddress, float Value)
{
    if (registerStartaddress > (m_numRegisters - 2))
    {
        return false;
    }
    /*小端模式*/
    slavemutex.lock();
    modbus_set_float(Value, &mapping->tab_registers[registerStartaddress]);
    slavemutex.unlock();
    return true;
}


bool CManageModbus::setInputRegisterValue(int registerStartaddress, float Value)
{
    if (registerStartaddress > (m_numRegisters - 2))
    {
        return false;
    }
    /*小端模式*/
    slavemutex.lock();
    modbus_set_float(Value, &mapping->tab_input_registers[registerStartaddress]);
    slavemutex.unlock();
    return true;
}

void CManageModbus::closeModbus()
{
    closesocket(m_modbusSocket);
    modbus_mapping_free(mapping);
    modbus_close(ctx);
    modbus_free(ctx);
    ctx = nullptr;
}

void CManageModbus::stopThread()
{
    m_bStopThread = true;
}


void CManageModbus::doWork()
{
    while (!m_bStopThread)
    {
        if (m_initialized)
        {
            recieveMessages();
        }
        else
        {
            m_initialized = true;
        }
    }


    int i = 1;
}

/***************************************************************
 * @file       CManageModbus.cpp
 * @author     seer-txj
 * @brief      获取寄存器里的浮点数 
 * @param      registerStartaddress寄存器起始地址
 * @version    v1
 * @return     两个uint16_t拼接而成的浮点值
 * @date       2021/10/6
 **************************************************************/
float CManageModbus::getHoldingRegisterFloatValue(int registerStartaddress)
{
    if (!m_initialized)
    {
        return -1.0f;
    }
    return modbus_get_float_badc(&mapping->tab_registers[registerStartaddress]);
}
/***************************************************************
 * @file       CManageModbus.cpp
 * @author     seer-txj
 * @brief      支持多个master同时连接
 * @version    v1
 * @return     null
 * @date       2021/10/6
 **************************************************************/
void CManageModbus::recieveMessages()
{
    uint8_t query[MODBUS_TCP_MAX_ADU_LENGTH];
    int master_socket;
    int rc;
    fd_set refset;
    fd_set rdset;
    /* Maximum file descriptor number */
    int fdmax;
    /* Clear the reference set of socket */
    FD_ZERO(&refset);
    /* Add the server socket */
    FD_SET(m_modbusSocket, &refset);

    /* Keep track of the max file descriptor */
    fdmax = m_modbusSocket;
    int nCount = 0;
    QString msg;

    while( true ) 
    {
        nCount++;
        rdset = refset;
        if (select(fdmax+1, &rdset, NULL, NULL, NULL) == -1) 
        {
            perror("Server select() failure.");
            break;
        }

        /* Run through the existing connections looking for data to be
         * read */
        for (master_socket = 0; master_socket <= fdmax; master_socket++) 
        {
            if (!FD_ISSET(master_socket, &rdset)) 
            {
                continue;
            }


            //qDebug() << "fdmax = " << fdmax << "master_socket = " << master_socket;

            if (master_socket == m_modbusSocket) 
            {
                /* A client is asking a new connection */
                socklen_t addrlen;
                struct sockaddr_in clientaddr;
                int newfd;

                /* Handle new connections */
                addrlen = sizeof(clientaddr);
                memset(&clientaddr, 0, sizeof(clientaddr));
                newfd = accept(m_modbusSocket, (struct sockaddr *)&clientaddr, &addrlen);
                if (newfd == -1) 
                {
                    perror("Server accept() error");
                }
                else
                {
                    FD_SET(newfd, &refset);

                    if (newfd > fdmax) 
                    {
                        /* Keep track of the maximum */
                        fdmax = newfd;
                    }
                    printf("New connection from %s:%d on socket %d\n", inet_ntoa(clientaddr.sin_addr), clientaddr.sin_port, newfd);
                }
                msg = QString("New connection from %1:%2 on socket %3").arg(inet_ntoa(clientaddr.sin_addr)).arg(clientaddr.sin_port).arg(newfd);
                qDebug() << msg;
                emit sigAddLog(m_bServer, true, msg);
            }
            else
            {
                modbus_set_socket(ctx, master_socket);

                rc = modbus_receive(ctx, query);

                if (rc > 0) 
                {
                    msg = unchar2QString(query, rc);
                    emit sigAddLog(m_bServer, true, msg);//接收到的数据显示到列表

                    if(m_bRemoveHead)
                    {
                        addMBAPhead(query, rc);//添加头
                    }

//                    QString str = "01 02 05 00 00 00 00 00 A2 52";
//                    uint8_t* send = new uint8_t[str.size()];
//                    int len = QStringHex2Char(str, send, 1);
                    int nSendLen = 0;
                    uint8_t *pSend = new uint8_t[260];
                    memset(pSend, 0, 260);

                    int errcode = modbus_reply(ctx, query, rc, mapping, pSend, &nSendLen, m_bRemoveHead?1:0);
                    if(errcode == -1)
                    {
                        msg = "发送失败";
                    }
                    else
                    {
                        msg = unchar2QString(pSend, nSendLen);
                    }
                    delete[] pSend;
                    pSend = nullptr;

                    emit sigAddLog(m_bServer, false, msg);

                }
                else if (rc == -1)
                {
                    /* This example server in ended on connection closing or
                     * any errors. */
                    /* 20240111屏蔽
                    msg = QString("Connection closed on socket %1").arg(master_socket);
                    emit sigAddLog(m_bServer, false, msg);
#ifdef _WIN32
                    closesocket(master_socket);
#else
                    close(master_socket);
#endif
                    //Remove from reference set
                    FD_CLR(master_socket, &refset);

                    if (master_socket == fdmax)
                    {
                        fdmax--;
                    }
                    */
                }
            }
        }
    }
    qDebug() << "out recieve";
    m_initialized = false;
}


QString CManageModbus::unchar2QString(unsigned char *pIn, int nLen)
{
    QString temp;
    QString msg;
    int j = 0;
    while(j < nLen)
    {
       temp = QString("%1 ").arg((int)pIn[j], 2, 16, QLatin1Char('0'));
       msg.append(temp);
       j++;
    }
    return msg;
}

int CManageModbus::QStringHex2Char(const QString &strIn, uint8_t *out, int lenSpace)
{
    uint8_t* p = nullptr;
    if (strIn.length() < 2)
    {
        return 0;
    }
    int offset = lenSpace + 2;//每次偏移2+间隔字符长度
    int j = 0;
    for (int i = 0; i < strIn.length(); i += offset)
    {
        QString num = strIn.mid(i, 2);
        bool ok = false;
        out[j] = num.toUInt(&ok, 16);
        j++;
        if (!ok)
        {
            return 0;
        }
    }
    return j;
}

void CManageModbus::addMBAPhead(uint8_t *query, int& rc)
{
    memcpy(query + 6, query, rc);
    memset(query, 0, 6);
    rc += 6;
}

void CManageModbus::removeMBAPhead(uint8_t *query, int &rc)
{
    memcpy(query, query + 6, rc - 6);
    rc -= 6;
}

void CManageModbus::buildSendMapping(const uint8_t *req, int req_len)
{

}

void CManageModbus::getTableData(int nDataType, uint8_t *pData, int nDataLen)
{
    switch (nDataType) {
    case TypeYxData:
        buildYxMapping(pData, nDataLen);
        break;
    case TypeYcData:
        //buildYcMapping(pData, nDataLen);
        break;
    case TypeYtData:
        buildYtMapping(pData, nDataLen);
        break;
    case TypeYkData:
        buildYkMapping(pData, nDataLen);
        break;
    case TypeYmData:
        buildYxMapping(pData, nDataLen);
        break;
    default:
        break;
    }
}

void CManageModbus::getConfigInfo(S_ModbusConfigInfo &configInfo)
{
    modbus_set_slave_id(configInfo.slaveId);
    m_configInfo = configInfo;
}

void CManageModbus::getTableData(int nDataType, uint16_t *pData, int nDataLen)
{
    switch (nDataType) {
    case TypeYxData:
        //buildYxMapping(pData, nDataLen);
        break;
    case TypeYcData:
        buildYcMapping(pData, nDataLen);
        break;
    case TypeYtData:
        //buildYtMapping(pData, nDataLen);
        break;
    case TypeYkData:
        //buildYkMapping(pData, nDataLen);
        break;
    case TypeYmData:
        //buildYxMapping(pData, nDataLen);
        break;
    default:
        break;
    }
}

void CManageModbus::buildYcMapping(uint16_t *pData, int nDataLen)
{
    if(mapping != nullptr)
    {
        memcpy(mapping->tab_registers + m_configInfo.ycStartAddr, pData, nDataLen * 2);//由于是uint16，所以得*2
    }
    int i = 1;
}

void CManageModbus::buildYxMapping(uint8_t *pData, int nDataLen)
{
    if(mapping != nullptr)
    {
        memcpy(mapping->tab_input_bits + m_configInfo.yxStartAddr, pData, nDataLen);
    }
}

void CManageModbus::buildYkMapping(uint8_t *pData, int nDataLen)
{
    if(mapping != nullptr)
    {
        memcpy(mapping->tab_registers, pData, nDataLen);
    }
}

void CManageModbus::buildYtMapping(uint8_t *pData, int nDataLen)
{
    if(mapping != nullptr)
    {
        memcpy(mapping->tab_registers, pData, nDataLen);
    }
}

void CManageModbus::buildYmMapping(uint8_t *pData, int nDataLen)
{
    if(mapping != nullptr)
    {
        memcpy(mapping->tab_registers, pData, nDataLen);
    }
}

unsigned int CManageModbus::ModBusCRC16(unsigned char *data, unsigned int len)
{
    unsigned int i, j, tmp, CRC16;

    CRC16 = 0xFFFF;             //CRC寄存器初始值
    for (i = 0; i < len; i++)
    {
        CRC16 ^= data[i];
        for (j = 0; j < 8; j++)
        {
            tmp = (unsigned int)(CRC16 & 0x0001);
            CRC16 >>= 1;
            if (tmp == 1)
            {
                CRC16 ^= 0xA001;    //异或多项式
            }
        }
    }
/*根据需要对结果进行处理*/
//    data[i++] = (unsigned char) (CRC16 & 0x00FF);
//    data[i++] = (unsigned char) ((CRC16 & 0xFF00)>>8);
    return CRC16;
}


